import pytest
from krbticket import KrbConfig
from datetime import timedelta
import os

DEFAULT_PRINCIPAL = 'user@EXAMPLE.COM'
DEFAULT_KEYTAB = './tests/conf/krb5.keytab'
DEFAULT_TICKET_RENEWAL_THRESHOLD_SEC = 1
DEFAULT_TICKET_LIFETIME_SEC = 4
DEFAULT_TICKET_LIFETIME = '{}s'.format(DEFAULT_TICKET_LIFETIME_SEC)
DEFAULT_TICKET_RENEWABLE_LIFETIME_SEC = 8
DEFAULT_TICKET_RENEWABLE_LIFETIME = '{}s'.format(DEFAULT_TICKET_RENEWABLE_LIFETIME_SEC)
DEFAULT_CCACHE_NAME = '/tmp/krb5cc_{}'.format(os.getuid())


def assert_ticket(t1, t2):
    assert t1.principal == t2.principal
    assert t1.file == t2.file
    assert t1.starting == t2.starting
    assert t1.expires == t2.expires
    assert t1.service_principal == t2.service_principal
    assert_config(t1.config, t2.config)


def assert_config(c1, c2):
    print(c1)
    print(c2)
    assert c1.principal == c2.principal
    assert c1.keytab == c2.keytab
    assert c1.kinit_bin == c2.kinit_bin
    assert c1.klist_bin == c2.klist_bin
    assert c1.kdestroy_bin == c2.kdestroy_bin
    assert c1.renewal_threshold == c2.renewal_threshold
    assert c1.ticket_lifetime == c2.ticket_lifetime
    assert c1.ticket_renewable_lifetime == c2.ticket_renewable_lifetime
    assert c1.ccache_name == c2.ccache_name


def default_config(**kwargs):
    default_options = {
        'principal': DEFAULT_PRINCIPAL,
        'keytab': DEFAULT_KEYTAB,
        'renewal_threshold': timedelta(seconds=DEFAULT_TICKET_RENEWAL_THRESHOLD_SEC),
        'ticket_lifetime': DEFAULT_TICKET_LIFETIME,
        'ticket_renewable_lifetime': DEFAULT_TICKET_RENEWABLE_LIFETIME,
        'retry_options': {
            'wait_exponential_multiplier': 100,
            'wait_exponential_max': 1000,
            'stop_max_attempt_number': 3}
        }
    return KrbConfig(**{**default_options, **kwargs})


@pytest.fixture
def config():
    return default_config()



import logging
import threading
import time

import fasteners

logger = logging.getLogger(__name__)


class KrbTicketUpdater(threading.Thread):
    DEFAULT_INTERVAL = 60 * 10

    def __init__(self, ticket, interval=DEFAULT_INTERVAL):
        super(KrbTicketUpdater, self).__init__()

        self.ticket = ticket
        self.interval = interval
        self.stop_event = threading.Event()
        self.daemon = True
        self.start_lock = threading.Lock()

    def run(self):
        logger.info("{} start...".format(self.__class__.__name__))
        while True:
            if self.stop_event.is_set():
                return

            logger.debug("Trying to update ticket...")
            self.ticket.maybe_update()
            time.sleep(self.interval)

    def start(self):
        with self.start_lock:
            if self.is_alive():
                logger.debug("Skipping Thread.start() since it already started...")
                return

            if self.stop_event.is_set():
                logger.debug("Skipping Thread.start() since it already stopped...")
                return

            super().start()

    @staticmethod
    def use_per_process_ccache():
        raise NotImplementedError

    def stop(self):
        logger.debug("Stopping ticket updater...")
        self.stop_event.set()


class SimpleKrbTicketUpdater(KrbTicketUpdater):
    """
    KrbTicketUpdater w/o exclusion control

    Using this with multiprocessing, child processes uses dedicated ccache file
    """
    @staticmethod
    def use_per_process_ccache():
        return True


class MultiProcessKrbTicketUpdater(KrbTicketUpdater):
    """
    Multiprocess KrbTicket Updater

    KrbTicketUpdater w/ exclusive lock for a ccache
    """
    @staticmethod
    def use_per_process_ccache():
        return False


class SingleProcessKrbTicketUpdater(KrbTicketUpdater):
    """
    Singleprocess KrbTicket Updater

    Single Process KrbTicketUpdater on the system.
    Multiple updaters can start, but they immediately stops if a updater is already running on the system.
    """
    def start(self):
        lock = fasteners.InterProcessLock(self.ticket.config.ccache_lockfile)
        got_lock = lock.acquire(blocking=False)
        if not got_lock:
            logger.debug("Another updater is detected. Stopping ticket updater...")
            return

        logger.debug("Got lock: {}...".format(self.ticket.config.ccache_lockfile))
        try:
            super().start()
        finally:
            if got_lock:
                lock.release()
                logger.debug("Released lock: {}...".format(self.ticket.config.ccache_lockfile))

    @staticmethod
    def use_per_process_ccache():
        return False
